/*
 * _KERMITS.C
 *
 * Contains:    _KermitSend()
 *              _KermitSendFile()
 *              _KermitSendFileName()
 *              _KermitFillBuffer()
 *              _KermitSendPacketAndWaitForACK()
 *
 *
 * The Greenleaf Comm Library
 *
 * Copyright (C) 1989-90 Greenleaf Software Inc.  All Rights Reserved.
 *
 */

#include <stdio.h>
#include <string.h>
#include "asiports.h"
#include "xfer.h"
#include "_xfer.h"

/*
 * Local prototypes for static procedures
 */

static int GF_CONV _KermitSendFile( XFER *kermit );
static int GF_CONV _KermitSendFileName( XFER *kermit );
static int GF_CONV _KermitFillBuffer( XFER *kermit, char *buffer );
static int GF_CONV _KermitSendPacketAndWaitForACK( XFER *kermit,
                                           char type,
                                           int buffer_length,
                                           char *buffer );

/*
 * int _KermitSend( XFER *kermit )
 *
 * ARGUMENTS
 *
 *  XFER *kermit:                      This is a pointer to an XFER block that
 *                                     keeps track of how things are set up
 *                                     for the file transfer.
 *
 * DESCRIPTION
 *
 * This routine is the private driver for performing a kermit file send.
 * It is called by the public setup routine, KermitSend(), which just
 * sets up the parameter block and then calls this guy.  The only
 * hitch in the whole deal is that the filename pointer passed to this
 * guy may actually be a list of files seperated by commas or spaces or
 * something like that.  So this routine picks individual file names
 * out of the list, one by one.  Some day I am thinking about using
 * strtok to pull them out, that might work better.  Anyway, the routine
 * has to first trade initialization packets with the far end.  After
 * that it sends the files, one by one, until they are all transferred.
 * It then sends the break packet.
 *
 * For the record, kermit packet types have the following definitions:
 *
 *   S:   The S-Init packet.  One of these comes at the startup phase
 *        and tells us a few things about
 *
 *   F:   This is a file header packet.  It tells me the name of the
 *        file that is about to be sent.
 *
 *   D:   This is a data packet.  Presumably it has stuff that is supposed
 *        to go into the file I have just opened.
 *
 *   Z:   This is an EOF packet.  It means I can close up the file.
 *
 *   B:   This is a break packet.  It means I can now take the connection
 *        down.
 *
 *   E:   This is an error packet.  It contains error text from the
 *        remote end.
 *
 *   Y:   This is an ACK packet.  It usually contains no data, but
 *        an ACK in response to an S packet should contain S data.
 *
 *   N:   This is a NAK packet.  It contains no data.
 *
 *   T:   This is a timeout packet.  A packet didn't really come, it
 *        means the receiver couldn't get any data.
 *
 *   Q:   This is a questionable packet.  A valid packet didn't really
 *        come through.  Something was wrong with it, like a bad checksum.
 *
 *   A:   This is an abort packet.  Usually happens when the user hits
 *        the abort key while reading data.
 *
 * SIDE EFFECTS
 *
 * File(s) may have been transfered.
 *
 * RETURNS
 *
 * The routine returns one of the defined XFER_RETURN codes defined
 * in XFER.H.
 *
 * AUTHOR
 *  Mark Nelson          28-Aug-1989  20:57:40.92
 *
 * MODIFICATIONS
 *
 */

int GF_CONV _KermitSend( XFER *kermit )
{
    char filename[ 81 ];   /* This holds each file name while I XFER it    */
    char *filename_list;   /* This points to the list I got when called.   */
    int i;

    kermit->sending=TRUE;
    if ( !_XferInitialize( kermit ) )
        return( FALSE );
    _XferMessage( kermit, "Sending S-init packet" );
/*
 * Here I store the pointer to the list of filenames, and set up the block
 * pointer to point to a local array in here.
 */
    filename_list = kermit->filename;
    kermit->filename = filename;
/*
 * Now I send and init packet, and try to get a packet back.
 */
    if ( !_KermitSendPacketAndWaitForACK( kermit,'S',6,_KermitInitParameters() ) )
        return( FALSE );
    _KermitLoadInitParameters( kermit );

    while ( TRUE )
    {
/*
 * This is the ugly part.  I look for a filename terminate by any one of
 * several characters.  I really ought to change this to use strtok()
 */
        i = 0;
        while ( *filename_list != ';' && *filename_list != ',' &&
                *filename_list != ' ' && *filename_list != '\0')
                 filename[ i++ ] = *filename_list++;
        if ( *filename_list == ';' || *filename_list == ',' ||
             *filename_list == ' ')
            filename_list++;
        filename[ i ] = '\0';

        if ( i != 0 ) {
            kermit->byte_count = 0L;
            if ( _XferAbortKeyPressed( kermit ) )
                return( FALSE );
            if ( !_KermitSendFile( kermit ) )
                return( FALSE );
        } else {
            break;
        }
    }
    return( _KermitSendPacketAndWaitForACK( kermit, 'B', 0, "" ) );
}

/*
 * static int _KermitSendFile( XFER *kermit )
 *
 * ARGUMENTS
 *
 *  XFER *kermit:                      This is a pointer to an XFER block that
 *                                     keeps track of how things are set up
 *                                     for the file transfer.
 *
 * DESCRIPTION
 *
 * This is a private routine called by _KermitSend.  It's job is to send
 * the single file pointed to by the filename pointer in the XFER block.
 * Since the init messages have already been transferred, all this guy
 * has to do is send the F packet with the filename, then send a whole
 * mess of D packets, followed by a Z packet.  Of course, the file has to
 * opened and closed.  One last detail:  I don't send the whole file name,
 * just the name and extension.  I toss out the drive and pathname.
 *
 *
 * SIDE EFFECTS
 *
 * A file may have been transfered.
 *
 * RETURNS
 *
 * The routine returns a TRUE if the file transfer went okay and we
 * can proceed, or FALSE if it was a disaster and it is time to abort.
 *
 * AUTHOR
 *  Mark Nelson          28-Sep-1989  20:57:40.92
 *
 * MODIFICATIONS
 *
 */

static int GF_CONV _KermitSendFile( XFER *kermit )
{
    char buffer[ 94 ];
    int buffer_size;

    _XferMessage( kermit, "Sending file %s", kermit->filename );
    if ( !_XferOpenFile( kermit ) )
        return( FALSE );
    if ( !_KermitSendFileName( kermit ) )
        return( FALSE );
    while ( TRUE ) {
        buffer_size = _KermitFillBuffer( kermit, buffer );
        if ( buffer_size == 0 )
            break;
        if ( !_KermitSendPacketAndWaitForACK( kermit, 'D', buffer_size, buffer) )
            return( FALSE );
        _XferMessage( kermit, "Sending packet %d", kermit->current_block_number );
    }
    fclose( kermit->file );
    _XferMessage( kermit, "Sending end of file packet" );
    if ( !_KermitSendPacketAndWaitForACK( kermit, 'Z', 0, "") )
        return( FALSE );
    else
        return( TRUE );
}

/*
 * static int _KermitSendFileName( XFER *kermit )
 *
 * ARGUMENTS
 *
 *  XFER *kermit:                      This is a pointer to an XFER block that
 *                                     keeps track of how things are set up
 *                                     for the file transfer.
 *
 * DESCRIPTION
 *
 * This is a private routine called by _KermitSendFile.  It has the easy
 * job of stripping the drive and path name from the file name, then
 * sending it to the remote end.  Pretty easy.
 *
 *
 * SIDE EFFECTS
 *
 *
 * RETURNS
 *
 * This routine returns a TRUE if everything went okay and we can proceed,
 * or a FALSE if things have gone bad and we need to abort.
 *
 * AUTHOR
 *  Mark Nelson          28-Sep-1989  20:57:40.92
 *
 * MODIFICATIONS
 *
 */

static int GF_CONV _KermitSendFileName( XFER *kermit )
{
    char *filename;

    filename = strrchr( kermit->filename, '\\' );
    if ( filename == NULL )
        filename = strrchr( kermit->filename, ':' );
    if ( filename == NULL )
        filename = kermit->filename - 1;
    filename++;
    return( _KermitSendPacketAndWaitForACK( kermit, 'F', strlen( filename ),
                                            filename ) );
}

/*
 * static int _KermitFillBuffer( XFER *kermit, char *buffer )
 *
 * ARGUMENTS
 *
 *  XFER *kermit:                      This is a pointer to an XFER block that
 *                                     keeps track of how things are set up
 *                                     for the file transfer.
 *
 * DESCRIPTION
 *
 * This is a private routine called by _KermitSendFile.  It has to fill
 * the buffer with a bunch of data.  The amount of data is determined by
 * the buffer side the remote end sent during our initialization phase.
 * The only reason this routine is a little complicated is that KERMIT
 * requires that control characters be quoted.  Note that if eightth
 * bit quoting was to be implemented on the send side, this is where it
 * would be done.
 *
 * SIDE EFFECTS
 *
 * RETURNS
 *
 * This routine returns a buffer count.  If the count is 0 it means
 * an EOF was reached.
 *
 * AUTHOR
 *  Mark Nelson          28-Sep-1989  20:57:40.92
 *
 * MODIFICATIONS
 *
 */

static int GF_CONV _KermitFillBuffer( XFER *kermit, char *buffer )
{
    int i;
    int c;

    i = 0;
    while ( i < ( kermit->x.kermit.max_length - 5 ) ) {
        c = getc( kermit->file );
        if ( c == EOF )
            break;
        kermit->byte_count++;
        if ( ( c & 0x7f ) < 0x20 || ( c & 0x7f ) == 0x7f ) {
            buffer[ i++ ] = KERMIT_MY_QUOTE_CHAR;
            buffer[ i++ ] = ctl( ( char )c );
        } else {
            if ( c == KERMIT_MY_QUOTE_CHAR )
                buffer[ i++ ] = KERMIT_MY_QUOTE_CHAR;
            buffer[ i++ ] = (char)c;
        }
    }
    return( i );
}


/*
 * static int _KermitSendPacketAndWaitForACK( XFER *kermit,
 *                                         char type,
 *                                         int buffer_length,
 *                                         char *buffer )
 *
 * ARGUMENTS
 *
 *  XFER *kermit:                      This is a pointer to an XFER block that
 *                                     keeps track of how things are set up
 *                                     for the file transfer.
 *
 *  char type:                         The type of packet.  See the packet
 *                                     descriptions at the top of this
 *                                     module for a better description.
 *
 *  int buffer_length:                 The number of bytes in the buffer
 *                                     that are ready to go out.
 *
 *  char *buffer:                      The data buffer.
 *
 * DESCRIPTION
 *
 * This is a private routine used during the sending of a file.  Its job
 * is to send a packet out and wait for a response from the remote end.
 * The only appropriate response would be an ACK, but all sorts of other
 * stuff may come back, and this routine decides what to do with the
 * other choices.  By the time this routine returns, it has either had
 * the buffer ACKed to its satisfaction, or their is just no point in
 * going on.
 *
 * SIDE EFFECTS
 *
 * A single buffer has been sent.
 *
 * RETURNS
 *
 * This routine returns a TRUE if everything went okay and we can proceed,
 * or a FALSE if things have gone bad and we need to abort.
 *
 * AUTHOR
 *  Mark Nelson          28-Sep-1989  20:57:40.92
 *
 * MODIFICATIONS
 *
 */

static int GF_CONV _KermitSendPacketAndWaitForACK( XFER *kermit,
                                           char type,
                                           int buffer_length,
                                           char *buffer )
{
    int return_type;
/*
 * This routine operates out of a forever loop, waiting for success or failure.
 */
    for ( ; ; ) {
/*
 * First send the packet out.  A failure to send is cause for abortion.
 */
        if ( !_KermitSendPacket( kermit, type, buffer_length, buffer ) )
            return( FALSE );
/*
 * Now I get a packet back, and take action based on the response type.
 */
        return_type = _KermitReceivePacket( kermit );
        switch ( return_type ) {
/*
 * An ACK is great, return success.
 */
            case 'Y' :
                if ( ( kermit->block_number & 0x3f ) ==
                       kermit->current_block_number ) {
                    kermit->block_number++;
                    return( TRUE );
                }
                _XferMessage( kermit, "Received a bad ACK packet number" );
                if ( !_KermitUpdateErrorCount( kermit ) )
                    return( FALSE );
                break;
/*
 * A NAK means send the packet again, if I haven't had too many errors.
 */
            case 'N' :
                _XferMessage( kermit, "Received a Nak packet" );
                if ( !_KermitUpdateErrorCount( kermit ) )
                    return( FALSE );
                break;
/*
 * An 'A' packet results from a user hitting the abort key.
 */
            case 'A' :
                return( FALSE );
/*
 * A 'T' packet means a timeout.  Make sure there haven't been too many
 * errors, and try again.
 */
            case 'T' :
                _XferMessage( kermit, "Timed out waiting for packet" );
                if ( !_KermitUpdateErrorCount( kermit ) )
                    return( FALSE );
                break;
/*
 * 'E' means the remote end sent an error code.  I quit when I get one.
 */
            case 'E' :
                _XferMessage( kermit, "Aborted from remote end" );
                kermit->return_status = XFER_RETURN_REMOTE_ABORT;
                _XferCleanup( kermit );
                return( FALSE );

            default :
                _XferMessage( kermit, "Funny packet" );
                if ( !_KermitUpdateErrorCount( kermit ) )
                    return( FALSE );
                break;
        }
    }
}

